@@ -138,13 +138,6 @@ static bool bch2_target_congested(struct bch_fs *c, u16 target)
 
 /* Cache promotion on read */
 
-static const struct rhashtable_params bch_promote_params = {
-	.head_offset		= offsetof(struct promote_op, hash),
-	.key_offset		= offsetof(struct promote_op, pos),
-	.key_len		= sizeof(struct bpos),
-	.automatic_shrinking	= true,
-};
-
 static inline bool have_io_error(struct bch_io_failures *failed)
 {
 	return failed && failed->nr;
@@ -168,7 +161,7 @@ static bool ptr_being_rewritten(struct bch_fs *c, struct bch_read_bio *orig, uns
 	struct bkey_ptrs_c ptrs = bch2_bkey_ptrs_c(bkey_i_to_s_c(u->k.k));
 	unsigned ptr_bit = 1;
 	bkey_for_each_ptr(ptrs, ptr) {
-		if (ptr->dev == dev && (u->opts.ptrs_rewrite & ptr_bit))
+		if (ptr->dev == dev && (u->opts.ptrs_kill & ptr_bit))
 			return true;
 		ptr_bit <<= 1;
 	}
@@ -177,34 +170,24 @@ static bool ptr_being_rewritten(struct bch_fs *c, struct bch_read_bio *orig, uns
 }
 
 static inline int should_promote(struct bch_fs *c, struct bkey_s_c k,
-				 struct bpos pos,
 				 struct bch_inode_opts opts,
-				 unsigned flags,
-				 bool self_healing)
+				 enum bch_read_flags flags)
 {
-	if (!self_healing) {
-		BUG_ON(!opts.promote_target);
+	BUG_ON(!opts.promote_target);
 
-		if (bch2_bkey_has_target(c, k, opts.promote_target)) {
-			event_inc(c, data_read_nopromote_already_promoted);
-			return bch_err_throw(c, nopromote_already_promoted);
-		}
-
-		if (bkey_extent_is_unwritten(c, k)) {
-			event_inc(c, data_read_nopromote_unwritten);
-			return bch_err_throw(c, nopromote_unwritten);
-		}
-
-		if (bch2_target_congested(c, opts.promote_target)) {
-			event_inc(c, data_read_nopromote_congested);
-			return bch_err_throw(c, nopromote_congested);
-		}
+	if (bch2_bkey_has_target(c, k, opts.promote_target)) {
+		event_inc(c, data_read_nopromote_already_promoted);
+		return bch_err_throw(c, nopromote_already_promoted);
 	}
 
-	if (rhashtable_lookup_fast(&c->promote_table, &pos,
-				   bch_promote_params)) {
-		event_inc(c, data_read_nopromote_in_flight);
-		return bch_err_throw(c, nopromote_in_flight);
+	if (bkey_extent_is_unwritten(c, k)) {
+		event_inc(c, data_read_nopromote_unwritten);
+		return bch_err_throw(c, nopromote_unwritten);
+	}
+
+	if (bch2_target_congested(c, opts.promote_target)) {
+		event_inc(c, data_read_nopromote_congested);
+		return bch_err_throw(c, nopromote_congested);
 	}
 
 	return 0;
@@ -215,10 +198,6 @@ static noinline void promote_free(struct bch_read_bio *rbio, int ret)
 	struct promote_op *op = container_of(rbio, struct promote_op, write.rbio);
 	struct bch_fs *c = rbio->c;
 
-	int ret2 = rhashtable_remove_fast(&c->promote_table, &op->hash,
-					  bch_promote_params);
-	BUG_ON(ret2);
-
 	async_object_list_del(c, promote, op->list_idx);
 	async_object_list_del(c, rbio, rbio->list_idx);
 
@@ -227,7 +206,7 @@ static noinline void promote_free(struct bch_read_bio *rbio, int ret)
 	bch2_data_update_exit(&op->write, ret);
 
 	enumerated_ref_put(&c->writes, BCH_WRITE_REF_promote);
-	kfree_rcu(op, rcu);
+	kfree_rcu(op, write.rcu);
 }
 
 static void promote_done(struct bch_write_op *wop)
@@ -264,14 +243,16 @@ static struct bch_read_bio *__promote_alloc(struct btree_trans *trans,
 					    struct bkey_s_c k,
 					    struct bpos pos,
 					    struct extent_ptr_decoded *pick,
-					    unsigned flags,
+					    enum bch_read_flags flags,
 					    unsigned sectors,
 					    struct bch_read_bio *orig,
 					    struct bch_io_failures *failed)
 {
 	struct bch_fs *c = trans->c;
 
-	int ret = should_promote(c, k, pos, orig->opts, flags, failed != NULL);
+	int ret = !failed
+		? should_promote(c, k, orig->opts, flags)
+		: 0;
 	if (ret)
 		return ERR_PTR(ret);
 
@@ -292,13 +273,13 @@ static struct bch_read_bio *__promote_alloc(struct btree_trans *trans,
 		bkey_for_each_ptr(ptrs, ptr) {
 			if (bch2_dev_io_failures(failed, ptr->dev) &&
 			    !ptr_being_rewritten(c, orig, ptr->dev)) {
-				update_opts.ptrs_io_error|= ptr_bit;
-				update_opts.ptrs_rewrite|= ptr_bit;
+				update_opts.ptrs_io_error |= ptr_bit;
+				update_opts.ptrs_kill |= ptr_bit;
 			}
 			ptr_bit <<= 1;
 		}
 
-		if (!update_opts.ptrs_rewrite)
+		if (!update_opts.ptrs_kill)
 			return ERR_PTR(bch_err_throw(c, nopromote_no_rewrites));
 	}
 
@@ -318,18 +299,11 @@ static struct bch_read_bio *__promote_alloc(struct btree_trans *trans,
 	}
 
 	op->start_time = local_clock();
-	op->pos = pos;
 	op->cpu	= cpu;
 
-	if (rhashtable_lookup_insert_fast(&c->promote_table, &op->hash,
-					  bch_promote_params)) {
-		ret = bch_err_throw(c, nopromote_in_flight);
-		goto err;
-	}
-
 	ret = async_object_list_add(c, promote, op, &op->list_idx);
 	if (ret < 0)
-		goto err_remove_hash;
+		goto err;
 
 	ret = bch2_data_update_init(trans, NULL, NULL, &op->write,
 			writepoint_hashed((unsigned long) current),
@@ -351,13 +325,10 @@ static struct bch_read_bio *__promote_alloc(struct btree_trans *trans,
 	return &op->write.rbio;
 err_remove_list:
 	async_object_list_del(c, promote, op->list_idx);
-err_remove_hash:
-	BUG_ON(rhashtable_remove_fast(&c->promote_table, &op->hash,
-				      bch_promote_params));
 err:
 	bch2_bio_free_pages_pool(c, &op->write.op.wbio.bio);
 	/* We may have added to the rhashtable and thus need rcu freeing: */
-	kfree_rcu(op, rcu);
+	kfree_rcu(op, write.rcu);
 err_up_limit:
 	up(per_cpu_ptr(c->promote_limit, cpu));
 err_put:
@@ -370,7 +341,7 @@ static struct bch_read_bio *promote_alloc(struct btree_trans *trans,
 					struct bvec_iter iter,
 					struct bkey_s_c k,
 					struct extent_ptr_decoded *pick,
-					unsigned flags,
+					enum bch_read_flags flags,
 					struct bch_read_bio *orig,
 					bool *bounce,
 					bool *read_full,
@@ -450,17 +421,12 @@ void bch2_promote_op_to_text(struct printbuf *out,
 void bch2_read_err_msg_trans(struct btree_trans *trans, struct printbuf *out,
 			     struct bch_read_bio *rbio, struct bpos read_pos)
 {
+	prt_str(out, "data read error at ");
 	bch2_inum_offset_err_msg_trans(trans, out, rbio->subvol, read_pos);
 
 	if (rbio->data_update)
-		prt_str(out, "(internal move) ");
-}
-
-static void bch2_read_err_msg(struct bch_fs *c, struct printbuf *out,
-			      struct bch_read_bio *rbio, struct bpos read_pos)
-{
-	CLASS(btree_trans, trans)(c);
-	bch2_read_err_msg_trans(trans, out, rbio, read_pos);
+		prt_str(out, " (internal move) ");
+	prt_str(out, ": ");
 }
 
 enum rbio_context {
@@ -611,7 +577,7 @@ static noinline int bch2_read_retry_nodecode(struct btree_trans *trans,
 					struct bch_read_bio *rbio,
 					struct bvec_iter bvec_iter,
 					struct bch_io_failures *failed,
-					unsigned flags)
+					enum bch_read_flags flags)
 {
 	struct data_update *u = container_of(rbio, struct data_update, rbio);
 	int ret = 0;
@@ -667,7 +633,7 @@ static void bch2_rbio_retry(struct work_struct *work)
 		container_of(work, struct bch_read_bio, work);
 	struct bch_fs *c	= rbio->c;
 	struct bvec_iter iter	= rbio->bvec_iter;
-	unsigned flags		= rbio->flags;
+	enum bch_read_flags flags = rbio->flags;
 	subvol_inum inum = {
 		.subvol = rbio->subvol,
 		.inum	= rbio->read_pos.inode,
@@ -675,6 +641,8 @@ static void bch2_rbio_retry(struct work_struct *work)
 	struct bpos read_pos = rbio->read_pos;
 	struct bch_io_failures failed = { .nr = 0 };
 
+	flags &= ~BCH_READ_hard_require_read_device;
+
 	event_inc_trace(c, data_read_retry, buf,
 			bch2_read_bio_to_text_atomic(&buf, rbio));
 
@@ -727,7 +695,6 @@ static void bch2_rbio_retry(struct work_struct *work)
 
 			bch2_read_err_msg_trans(trans, &msg.m, rbio, read_pos);
 
-			prt_str(&msg.m, "data read error, ");
 			if (!ret) {
 				prt_str(&msg.m, "successful retry");
 				if (rbio->self_healing)
@@ -834,11 +801,17 @@ static noinline void bch2_rbio_narrow_crcs(struct bch_read_bio *rbio)
 		}));
 }
 
+static int bch2_rbio_decrypt(struct bch_fs *c, struct bch_read_bio *rbio,
+			     struct bch_extent_crc_unpacked crc, struct nonce nonce)
+{
+	return bch2_encrypt_bio(c, crc.csum_type, nonce, &rbio->bio)
+		? bch_err_throw(c, data_read_decrypt_err)
+		: 0;
+}
+
 /* Inner part that may run in process context */
-static void __bch2_read_endio(struct work_struct *work)
+static int __bch2_read_endio_work(struct bch_read_bio *rbio)
 {
-	struct bch_read_bio *rbio =
-		container_of(work, struct bch_read_bio, work);
 	struct bch_fs *c	= rbio->c;
 	struct bch_dev *ca = rbio->have_ioref ? bch2_dev_have_ref(c, rbio->pick.ptr.dev) : NULL;
 	struct bch_read_bio *parent	= bch2_rbio_parent(rbio);
@@ -875,16 +848,13 @@ static void __bch2_read_endio(struct work_struct *work)
 	 */
 	if (!csum_good && !rbio->bounce && (rbio->flags & BCH_READ_user_mapped)) {
 		rbio->flags |= BCH_READ_must_bounce;
-		bch2_rbio_error(rbio, bch_err_throw(c, data_read_retry_csum_err_maybe_userspace));
-		return;
+		return bch_err_throw(c, data_read_retry_csum_err_maybe_userspace);
 	}
 
 	bch2_account_io_completion(ca, BCH_MEMBER_ERROR_checksum, 0, csum_good);
 
-	if (!csum_good) {
-		bch2_rbio_error(rbio, bch_err_throw(c, data_read_retry_csum_err));
-		return;
-	}
+	if (!csum_good)
+		return bch_err_throw(c, data_read_retry_csum_err);
 
 	/*
 	 * XXX
@@ -902,17 +872,13 @@ static void __bch2_read_endio(struct work_struct *work)
 		crc.live_size	= bvec_iter_sectors(rbio->bvec_iter);
 
 		if (crc_is_compressed(crc)) {
-			ret = bch2_encrypt_bio(c, crc.csum_type, nonce, src);
-			if (ret) {
-				bch2_rbio_error(rbio, bch_err_throw(c, data_read_decrypt_err));
-				return;
-			}
+			BUG_ON(!rbio->bounce);
 
-			if (bch2_bio_uncompress(c, src, dst, dst_iter, crc) &&
-			    !c->opts.no_data_io) {
-				bch2_rbio_error(rbio, bch_err_throw(c, data_read_decompress_err));
-				return;
-			}
+			try(bch2_rbio_decrypt(c, rbio, crc, nonce));
+
+			ret = bch2_bio_uncompress(c, src, dst, dst_iter, crc);
+			if (ret && !c->opts.no_data_io)
+				return ret;
 		} else {
 			/* don't need to decrypt the entire bio: */
 			nonce = nonce_add(nonce, crc.offset << 9);
@@ -921,11 +887,7 @@ static void __bch2_read_endio(struct work_struct *work)
 			BUG_ON(src->bi_iter.bi_size < dst_iter.bi_size);
 			src->bi_iter.bi_size = dst_iter.bi_size;
 
-			ret = bch2_encrypt_bio(c, crc.csum_type, nonce, src);
-			if (ret) {
-				bch2_rbio_error(rbio, bch_err_throw(c, data_read_decrypt_err));
-				return;
-			}
+			try(bch2_rbio_decrypt(c, rbio, crc, nonce));
 
 			if (rbio->bounce) {
 				struct bvec_iter src_iter = src->bi_iter;
@@ -934,6 +896,27 @@ static void __bch2_read_endio(struct work_struct *work)
 			}
 		}
 	} else {
+		if (parent->data_update_verify_decompress &&
+		    crc_is_compressed(crc)) {
+			BUG_ON(!rbio->bounce);
+
+			try(bch2_rbio_decrypt(c, rbio, crc, nonce));
+
+			/*
+			 * dst_iter doesn't make sense here, it refers to the
+			 * size of the compressed extent on disk (what the data
+			 * update path generally wants), and here we're just
+			 * verifying thath the data decompresses and throwing
+			 * away the result:
+			 */
+			ret = bch2_bio_uncompress(c, src, dst, (struct bvec_iter) {}, crc);
+			if (ret && !c->opts.no_data_io)
+				return ret;
+
+			/* We decrypted to decompress; re-encrypt: */
+			try(bch2_rbio_decrypt(c, rbio, crc, nonce));
+		}
+
 		if (rbio->split)
 			rbio->parent->pick = rbio->pick;
 
@@ -949,17 +932,24 @@ static void __bch2_read_endio(struct work_struct *work)
 		 * Re encrypt data we decrypted, so it's consistent with
 		 * rbio->crc:
 		 */
-		ret = bch2_encrypt_bio(c, crc.csum_type, nonce, src);
-		if (ret) {
-			bch2_rbio_error(rbio, bch_err_throw(c, data_read_decrypt_err));
-			return;
-		}
+		try(bch2_rbio_decrypt(c, rbio, crc, nonce));
 	}
 
 	if (likely(!(rbio->flags & BCH_READ_in_retry))) {
 		rbio = bch2_rbio_free(rbio);
 		bch2_rbio_done(rbio);
 	}
+
+	return 0;
+}
+
+static void bch2_read_endio_work(struct work_struct *work)
+{
+	struct bch_read_bio *rbio =
+		container_of(work, struct bch_read_bio, work);
+	int ret = __bch2_read_endio_work(rbio);
+	if (ret)
+		bch2_rbio_error(rbio, ret);
 }
 
 static void bch2_read_endio(struct bio *bio)
@@ -1004,7 +994,7 @@ static void bch2_read_endio(struct bio *bio)
 	else if (rbio->pick.crc.csum_type)
 		context = RBIO_CONTEXT_HIGHPRI,	wq = system_highpri_wq;
 
-	bch2_rbio_punt(rbio, __bch2_read_endio, context, wq);
+	bch2_rbio_punt(rbio, bch2_read_endio_work, context, wq);
 }
 
 static noinline void read_from_stale_dirty_pointer(struct btree_trans *trans,
@@ -1062,7 +1052,7 @@ static inline struct bch_read_bio *read_extent_rbio_alloc(struct btree_trans *tr
 			struct extent_ptr_decoded pick,
 			struct bch_dev *ca,
 			unsigned offset_into_extent,
-			struct bch_io_failures *failed, unsigned flags,
+			struct bch_io_failures *failed, enum bch_read_flags flags,
 			bool bounce, bool read_full, bool narrow_crcs)
 {
 	struct bch_fs *c = trans->c;
@@ -1206,7 +1196,7 @@ static inline struct bch_read_bio *read_extent_rbio_alloc(struct btree_trans *tr
 	return rbio;
 }
 
-static inline int read_extent_done(struct bch_read_bio *rbio, unsigned flags, int ret)
+static inline int read_extent_done(struct bch_read_bio *rbio, enum bch_read_flags flags, int ret)
 {
 	if (flags & BCH_READ_in_retry)
 		return ret;
@@ -1224,7 +1214,7 @@ static noinline int read_extent_inline(struct bch_fs *c,
 				       struct bvec_iter iter,
 				       struct bkey_s_c k,
 				       unsigned offset_into_extent,
-				       unsigned flags)
+				       enum bch_read_flags flags)
 {
 	event_add_trace(c, data_read_inline, bvec_iter_sectors(iter), buf, ({
 		bch2_bkey_val_to_text(&buf, c, k);
@@ -1256,7 +1246,7 @@ static noinline int read_extent_hole(struct bch_fs *c,
 				     struct bch_read_bio *rbio,
 				     struct bvec_iter iter,
 				     struct bkey_s_c k,
-				     unsigned flags)
+				     enum bch_read_flags flags)
 {
 	event_add_trace(c, data_read_hole, bvec_iter_sectors(iter), buf, ({
 		bch2_bkey_val_to_text(&buf, c, k);
@@ -1281,7 +1271,7 @@ static noinline int read_extent_pick_err(struct btree_trans *trans,
 					 struct bch_read_bio *rbio,
 					 struct bpos read_pos,
 					 enum btree_id data_btree, struct bkey_s_c k,
-					 unsigned flags, int ret)
+					 enum bch_read_flags flags, int ret)
 {
 	struct bch_fs *c = trans->c;
 
@@ -1307,7 +1297,7 @@ static noinline int read_extent_no_encryption_key(struct btree_trans *trans,
 					 struct bch_read_bio *rbio,
 					 struct bpos read_pos,
 					 struct bkey_s_c k,
-					 unsigned flags)
+					 enum bch_read_flags flags)
 {
 	struct bch_fs *c = trans->c;
 
@@ -1326,12 +1316,12 @@ int __bch2_read_extent(struct btree_trans *trans,
 		       struct bvec_iter iter, struct bpos read_pos,
 		       enum btree_id data_btree, struct bkey_s_c k,
 		       unsigned offset_into_extent,
-		       struct bch_io_failures *failed, unsigned flags, int dev)
+		       struct bch_io_failures *failed,
+		       enum bch_read_flags flags, int dev)
 {
 	struct bch_fs *c = trans->c;
 	struct extent_ptr_decoded pick;
-	bool bounce = false, read_full = false, narrow_crcs = false;
-	struct data_update *u = rbio_data_update(orig);
+	bool bounce = false, read_full = false;
 	int ret = 0;
 
 	if (bkey_extent_is_inline_data(k.k))
@@ -1341,7 +1331,7 @@ int __bch2_read_extent(struct btree_trans *trans,
 	    !orig->data_update)
 		return read_extent_done(orig, flags, bch_err_throw(c, extent_poisoned));
 
-	ret = bch2_bkey_pick_read_device(c, k, failed, &pick, dev);
+	ret = bch2_bkey_pick_read_device(c, k, failed, &pick, dev, flags);
 
 	/* hole or reservation - just zero fill: */
 	if (unlikely(!ret))
@@ -1376,18 +1366,44 @@ int __bch2_read_extent(struct btree_trans *trans,
 		return read_extent_done(orig, flags, bch_err_throw(c, data_read_ptr_stale_dirty));
 	}
 
-	if (likely(!u)) {
-		if (!(flags & BCH_READ_last_fragment) ||
-		    bio_flagged(&orig->bio, BIO_CHAIN))
-			flags |= BCH_READ_must_clone;
+	if (!(flags & BCH_READ_last_fragment) ||
+	    bio_flagged(&orig->bio, BIO_CHAIN))
+		flags |= BCH_READ_must_clone;
 
-		narrow_crcs = !(flags & BCH_READ_in_retry) && can_narrow_crc(pick.crc);
+	bool narrow_crcs = !orig->data_update &&
+		!(flags & BCH_READ_in_retry) &&
+		can_narrow_crc(pick.crc);
 
-		if (narrow_crcs && (flags & BCH_READ_user_mapped))
-			flags |= BCH_READ_must_bounce;
+	if (narrow_crcs && (flags & BCH_READ_user_mapped))
+		flags |= BCH_READ_must_bounce;
 
+
+	if (likely(!orig->data_update)) {
+		/* Check we're not trying to read more than we have in this extent: */
 		EBUG_ON(offset_into_extent + bvec_iter_sectors(iter) > k.k->size);
+	} else {
+		/*
+		 * For data updates, iter.bi_size can be bigger than the key - it's just
+		 * the buffer size, we're not trying to read a specific amount of data
+		 * from this extent:
+		 *
+		 * If it's smaller, we raced with a merge during a retry:
+		 */
+		struct data_update *u = rbio_data_update(orig);
+		if (unlikely(pick.crc.uncompressed_size > u->op.wbio.bio.bi_iter.bi_size)) {
+			if (ca)
+				enumerated_ref_put(&ca->io_ref[READ],
+					BCH_DEV_READ_REF_io_read);
+			return read_extent_done(orig, flags, bch_err_throw(c, data_read_buffer_too_small));
+		}
 
+		iter.bi_size = pick.crc.compressed_size << 9;
+		read_full = true;
+	}
+
+	bool decode = !orig->data_update || orig->data_update_verify_decompress;
+
+	if (likely(decode))
 		if (crc_is_compressed(pick.crc) ||
 		    (pick.crc.csum_type != BCH_CSUM_none &&
 		     (bvec_iter_sectors(iter) != pick.crc.uncompressed_size ||
@@ -1397,21 +1413,6 @@ int __bch2_read_extent(struct btree_trans *trans,
 			read_full = true;
 			bounce = true;
 		}
-	} else {
-		/*
-		 * can happen if we retry, and the extent we were going to read
-		 * has been merged in the meantime:
-		 */
-		if (unlikely(pick.crc.compressed_size > u->op.wbio.bio.bi_iter.bi_size)) {
-			if (ca)
-				enumerated_ref_put(&ca->io_ref[READ],
-					BCH_DEV_READ_REF_io_read);
-			return read_extent_done(orig, flags, bch_err_throw(c, data_read_buffer_too_small));
-		}
-
-		iter.bi_size	= pick.crc.compressed_size << 9;
-		read_full = true;
-	}
 
 	struct bch_read_bio *rbio =
 		read_extent_rbio_alloc(trans, orig, iter, read_pos, data_btree, k,
@@ -1481,7 +1482,7 @@ int __bch2_read(struct btree_trans *trans, struct bch_read_bio *rbio,
 		struct bvec_iter bvec_iter, subvol_inum inum,
 		struct bch_io_failures *failed,
 		struct bkey_buf *prev_read,
-		unsigned flags)
+		enum bch_read_flags flags)
 {
 	struct bch_fs *c = trans->c;
 	struct bkey_s_c k;
@@ -1648,15 +1649,17 @@ void bch2_read_bio_to_text(struct printbuf *out,
 			   struct bch_fs *c,
 			   struct bch_read_bio *rbio)
 {
-	bch2_read_err_msg(c, out, rbio, rbio->read_pos);
+	CLASS(btree_trans, trans)(c);
+	bch2_inum_offset_err_msg_trans(trans, out, rbio->subvol, rbio->read_pos);
+
+	if (rbio->data_update)
+		prt_str(out, " (internal move) ");
 	prt_newline(out);
 	__bch2_read_bio_to_text(out, rbio);
 }
 
 void bch2_fs_io_read_exit(struct bch_fs *c)
 {
-	if (c->promote_table.tbl)
-		rhashtable_destroy(&c->promote_table);
 	bioset_exit(&c->bio_read_split);
 	bioset_exit(&c->bio_read);
 	mempool_exit(&c->bio_bounce_bufs);
@@ -1701,8 +1704,5 @@ int bch2_fs_io_read_init(struct bch_fs *c)
 			BIOSET_NEED_BVECS))
 		return bch_err_throw(c, ENOMEM_bio_read_split_init);
 
-	if (rhashtable_init(&c->promote_table, &bch_promote_params))
-		return bch_err_throw(c, ENOMEM_promote_table_init);
-
 	return 0;
 }
